import argparse
from . import job
import subprocess
import os
import re
import xyz_py as xyzp

from .utils import red_exit, get_opt_coords, get_input_section, blue_print


def gen_job_func(uargs):
    '''
    Wrapper for CLI gen_job call

    Parameters
    ----------
    uargs : argparser object
        User arguments

    Returns
    -------
    None

    '''

    # Currently available nodes
    supported_nodes = [
        'spot-fsv2-2',
        'spot-fsv2-4',
        'spot-fsv2-8',
        'spot-fsv2-16',
        'spot-fsv2-32',
        'paygo-fsv2-2',
        'paygo-fsv2-4',
        'paygo-fsv2-8',
        'paygo-fsv2-16',
        'paygo-fsv2-32',
        'paygo-hb-60',
        'paygo-hbv2-120',
        'paygo-hbv3-120',
        'paygo-hc-44',
        'paygo-ncv3-12',
        'paygo-ncv3-24',
        'paygo-ncv3-6',
        'paygo-ncv3r-24',
        'paygo-ndv2-40',
        'spot-hb-60',
        'spot-hbv2-120',
        'spot-hbv3-120',
        'spot-hc-44',
        'spot-ncv3-12',
        'spot-ncv3-24',
        'spot-ncv3-6',
        'spot-ncv3r-24',
        'spot-ndv2-40',
        'vis-ncv3-12',
        'vis-ncv3-24',
        'vis-ncv3-6',
        'vis-ndv2-40'
    ]

    cores_per_node = {
        node: int(node.split('-')[-1])
        for node in supported_nodes
    }

    total_node_memory = {
        'spot-fsv2-2': 4000,
        'spot-fsv2-4': 8000,
        'spot-fsv2-8': 16000,
        'spot-fsv2-16': 32000,
        'spot-fsv2-32': 64000,
        'paygo-fsv2-2': 4000,
        'paygo-fsv2-4': 8000,
        'paygo-fsv2-8': 16000,
        'paygo-fsv2-16': 32000,
        'paygo-fsv2-32': 64000,
        'paygo-hbv2-120': 456000,
        'paygo-hbv3-120': 448000,
        'paygo-hc-44': 352000,
        'paygo-ncv3-6': 112000,
        'paygo-ncv3-12': 224000,
        'paygo-ncv3-24': 448000,
        'paygo-ncv3r-24': 18500,
        'paygo-ndv2-40': 672000,
        'spot-hbv2-120': 456000,
        'spot-hbv3-120': 448000,
        'spot-hc-44': 352000,
        'spot-ncv3-6': 112000,
        'spot-ncv3-12': 224000,
        'spot-ncv3-24': 448000,
        'spot-ncv3r-24': 18500,
        'spot-ndv2-40': 672000,
        'vis-ncv3-6': 112000,
        'vis-ncv3-12': 224000,
        'vis-ncv3-24': 448000,
        'vis-ndv2-40': 672000
    }

    if uargs.node_type in supported_nodes:
        node = uargs.node_type
    else:
        red_exit("Node type unsupported")

    # Write job file
    for file in uargs.input_files:

        # Check input exists
        if not os.path.exists(file):
            red_exit("Cannot locate {}".format(file))

        # Check contents of input file and find any file dependencies
        _, input_deps_rel = job.parse_input_contents(
            file,
            total_node_memory[node],
            cores_per_node[node]
        )

        # Check for old results folder and look at contents
        # to see if any restart capable files exist
        result_deps = job.parse_results_contents(file)

        # Resolve different versions of same filetypes
        dependencies = job.resolve_deps(input_deps_rel, result_deps)

        job_file = job.write_file(
            file, node, uargs.time, verbose=True, dependencies=dependencies
        )

        # Submit to queue
        if not uargs.no_start:
            subprocess.call("sbatch {}".format(job_file), shell=True)

    return


def rst_opt_func(uargs, job_args):

    path, raw_file = os.path.split(uargs.output_file)
    head = os.path.splitext(raw_file)[0]

    # Extract coordinates from output file
    labels, coords, opt_yn = get_opt_coords(uargs.output_file)

    # Extract input information from output file
    input_info = get_input_section(uargs.output_file)

    # Create rst folder
    new_folder = os.path.join(path, 'rst')
    os.mkdir(new_folder)

    # Create -rst xyz file
    new_xyz = os.path.join(new_folder, "{}-rst.xyz".format(head))
    xyzp.save_xyz(new_xyz, labels, coords, verbose=False)

    # Edit xyz file name in input_info
    input_info = re.sub(
        r"[\-a-z0-9A-Z_]+\.xyz",
        "{}-rst.xyz".format(head),
        input_info
    )

    # If optimised, delete opt keyword from input
    if opt_yn:
        input_info = re.sub(
            r"\bopt\b(?!-)(?!\.)",
            "",
            input_info
        )
        blue_print("Optimisation complete, restarting only for frequencies")

    # Create -rst input file
    new_input = os.path.join(new_folder, "{}-rst.inp".format(head))
    with open(new_input, 'w') as f:
        f.write(input_info)

    # Run gen_job on new calculation
    read_args(
        [
            "gen_job",
            new_input,
            *job_args
        ]
    )

    return


def read_args(arg_list=None):
    '''
    Reader for command line arguments. Uses subReaders for individual programs

    Parameters
    ----------
        args : argparser object
            command line arguments

    Returns
    -------
        None

    '''

    description = '''
    A package for working with Orca on Bath's Cloud HPC service
    '''

    epilog = '''
    To display options for a specific program, use splash \
    PROGRAMFILETYPE -h
    '''
    parser = argparse.ArgumentParser(
        description=description,
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    subparsers = parser.add_subparsers(dest='prog')

    gen_job = subparsers.add_parser(
        'gen_job',
        description='Generate Nimbus SLURM submission script'
    )
    gen_job.set_defaults(func=gen_job_func)

    gen_job.add_argument(
        'input_files',
        nargs='+',
        type=str,
        help='Orca input file name(s)'
    )

    gen_job.add_argument(
        '-nt',
        '--node_type',
        default='spot-fsv2-16',
        type=str,
        help='Node to run on, default is spot-fsv2-16'
    )

    gen_job.add_argument(
        '-t',
        '--time',
        type=str,
        default='24:00:00',
        help='Time for job, formatted as HH:MM:SS, default 24:00:00'
    )

    gen_job.add_argument(
        '-ns',
        '--no_start',
        action='store_true',
        help='If specified, jobs are not submitted to nimbus queue'
    )

    rst_opt = subparsers.add_parser(
        'rst_opt',
        description='Restart optimisation from output file'
    )
    rst_opt.set_defaults(func=rst_opt_func)

    rst_opt.add_argument(
        'output_file',
        type=str,
        help='Orca output file name(s) (must contain coordinates from optimisation)' # noqa
    )

    # If argument list is none, then call function func
    # which is assigned to help function
    parser.set_defaults(func=lambda user_args: parser.print_help())

    # read sub-parser
    _args, _ = parser.parse_known_args(arg_list)

    # select parsing option based on sub-parser
    if _args.prog in ['rst_opt']:
        args, job_args = parser.parse_known_args(arg_list)
        args.func(args, job_args)
    else:
        args = parser.parse_args(arg_list)
        args.func(args)
    return args


def interface():
    read_args()
    return



import os
import sys
import subprocess

from .utils import red_exit


def write_file(input_file: str, node_type: str, time: str,
               verbose: bool = False, dependencies: list[str] = []) -> str:
    """
    Writes slurm jobscript to file for ORCA calculation on nimbus

    Output file name is input_file with .slm extension

    Parameters
    ----------
    input_file : str
        Name of input file, including extension
    node_type : str
        Name of Nimbus node to use
    time : str
        Job time limit formatted as HH:MM:SS
    verbose : bool, default=False
        If True, prints job file name to screen
    dependencies : list[str]
        Additional files on which this job depends, these will be copied
        to the compute node at runtime

    Returns
    -------
    str
        Name of jobscript file
    """

    # Check for research allocation id environment variable
    check_envvar('CLOUD_ACC')

    inpath, in_raw = os.path.split(input_file)

    # Name of job
    job_name = os.path.splitext(in_raw)[0]

    # Job file to write to, must use path if present in input_file
    job_file = os.path.join(
        inpath,
        "{}.slm".format(job_name)
    )

    # Path of jobfile, input, xyz, gbw etc
    if not len(inpath):
        calc_dir = os.path.abspath(os.getcwd())
    else:
        calc_dir = os.path.abspath(inpath)

    with open(job_file, 'w') as j:

        j.write('#!/bin/bash\n\n')

        j.write('#SBATCH --job-name={}\n'.format(job_name))
        j.write('#SBATCH --nodes=1\n')
        j.write('#SBATCH --ntasks-per-node={}\n'.format(
            node_type.split('-')[-1])
        )
        j.write('#SBATCH --partition={}\n'.format(node_type))
        j.write('#SBATCH --account={}\n'.format(os.environ['CLOUD_ACC']))
        j.write('#SBATCH --qos={}\n'.format(node_type))
        j.write('#SBATCH --output={}.%j.o\n'.format(job_name))
        j.write('#SBATCH --error={}.%j.e\n'.format(job_name))
        j.write('#SBATCH --signal=B:USR1\n\n')

        j.write('# Job time\n')
        j.write('#SBATCH --time={}\n\n'.format(time))

        j.write('# name and path of the input/output files and locations\n')
        j.write('input={}\n'.format(in_raw))
        j.write('output={}.out\n'.format(job_name))
        j.write('campaigndir={}\n'.format(calc_dir))
        j.write('results=$campaigndir/{}\n\n'.format(job_name))

        j.write('# If results directory already exists, append OLD and ')
        j.write('last access time\n')
        j.write('if [ -d $results ]; then\n')
        j.write(
            '    mv $results "$results"_OLD_$(date -r $results "+%m-%d-%Y")\n')
        j.write('fi\n\n')

        j.write('# If output file already exists, append OLD and ')
        j.write('last access time\n')
        j.write('if [ -d $output ]; then\n')
        j.write(
            '    mv $output "$output"_OLD_$(date -r $output "+%m-%d-%Y")\n')
        j.write('fi\n\n')

        j.write('# Local (Node) scratch, either node itself if supported')
        j.write('or burstbuffer\n')
        j.write('if [ -d "/mnt/resource/" ]; then\n')
        j.write(
            '    localscratch="/mnt/resource/temp_scratch_$SLURM_JOB_ID"\n'
            '    mkdir $localscratch\n'
        )
        j.write('else\n')
        j.write('    localscratch=$BURSTBUFFER\n')
        j.write('fi\n\n')

        j.write('# Copy files to localscratch\n')
        j.write("rsync -aP ")

        j.write('$campaigndir/{}'.format(in_raw))

        for dep in dependencies:
            j.write(' $campaigndir/{}'.format(dep))

        j.write(" $localscratch\n")
        j.write('cd $localscratch\n\n')

        j.write('# write date and node type to output\n')
        j.write('date > $campaigndir/$output\n')
        j.write('uname -n >> $campaigndir/$output\n\n')

        j.write('# Module system setup\n')
        j.write('source /apps/build/easy_build/scripts/id_instance.sh\n')
        j.write('source /apps/build/easy_build/scripts/setup_modules.sh\n\n')

        j.write('# Load orca\n')
        j.write('module purge\n')
        j.write('module load ORCA/5.0.1-gompi-2021a\n\n')

        j.write('# UCX transport protocols for MPI\n')
        j.write('export UCX_TLS=self,tcp,sm\n\n')

        j.write('# If timeout, evicted, cancelled, then manually end orca\n')

        j.write("trap 'echo signal recieved in BATCH!; kill -15 ")
        j.write('"${PID}"; wait "${PID}";')
        j.write("' SIGINT SIGTERM USR1 15\n\n")

        j.write('# Run calculation in background\n')
        j.write('# Catch the PID var for trap, and wait for process to end\n')
        j.write('$(which orca) $input >> $campaigndir/$output &\n')
        j.write('PID="$!"\n')
        j.write('wait "${PID}"\n\n')

        j.write('Clean up and copy back files\n')
        j.write('rsync -aP --exclude=*.tmp* $localscratch/* $results\n')
        j.write('rm -r $localscratch\n')

    if verbose:
        print("\u001b[32mSubmission script written to {} \033[0m".format(
            job_file
        ))

    return job_file


def check_envvar(var_str: str) -> None:
    """
    Checks specified environment variable has been defined, exits program if
    variable is not defined

    Parameters
    ----------
    var_str : str
        String name of environment variable

    Returns
    -------
    None
    """

    try:
        os.environ[var_str]
    except KeyError:
        sys.exit("Please set ${} environment variable".format(var_str))

    return


def parse_input_contents(input_file: str, max_mem: int,
                         max_cores: int) -> tuple[
        dict[str: str], dict[str: str]]:
    """
    Checks contents of input file and returns file dependencies
    Specific checks:
        If specified xyz file exists
        If specified gbw file exists
        If maxcore (memory) specified is appropriate

    Parameters
    ----------
    input_file : str
        Name of orca input file
    max_mem : int
        Max memory (MB) total on node
    max_cores : int
        Maximum number of cores on node

    Returns
    -------
    dict[str: str]
        Names of full-path dependencies (files) which this input needs
        key is identifier (xyz, gbw), value is file name
    dict[str: str]
        Names of relative-path dependencies (files) which this input needs
        key is identifier (xyz, gbw), value is file name
    """

    # Found memory and cores
    mem_found = False
    core_found = False

    # Dependencies (files) of this input file
    # as either full or relative path
    full_path_deps = dict()
    rel_path_deps = dict()

    # Path of input file if provided
    inpath = os.path.split(input_file)[0]

    with open(input_file, 'r') as f:
        for line in f:

            # xyz file
            if 'xyzfile' in line.lower():
                if len(line.split()) != 5 and line.split()[0] != '*xyzfile':
                    red_exit(
                        "Incorrect xyzfile definition in {}".format(input_file)
                    )

                if len(line.split()) != 4 and line.split()[0] != '*':
                    red_exit(
                        "Incorrect xyzfile definition in {}".format(input_file)
                    )

                xyzfile = line.split()[-1]

                if not os.path.exists(os.path.join(inpath, xyzfile)):
                    red_exit(
                        "xyz file specified in {} cannot be found".format(
                            input_file
                        )
                    )

                rel_path_deps["xyz"] = xyzfile

            # gbw file
            if '%moinp' in line.lower():
                if len(line.split()) != 2:
                    red_exit(
                        "Incorrect gbw_file definition in {}".format(
                            input_file
                        )
                    )

                gbw_file = line.split()[-1].replace('"', '').replace("'", "")

                # Absolute path
                if gbw_file == os.path.abspath(gbw_file):
                    full_path_deps["gbw"] = os.path.abspath(gbw_file)
                # Relative path (must be cwd)
                elif os.sep not in gbw_file:
                    rel_path_deps["gbw"] = gbw_file
                # Neither, then error!
                else:
                    red_exit(
                        "Path to gbw file specified in {} must be absolute or name of file in current directory".format(input_file) # noqa
                    )

                if os.path.basename(gbw_file) == os.path.basename(input_file):
                    red_exit(
                        "gbw file cannot have same base name as {}".format(
                            input_file
                        )
                    )

                if not os.path.exists(os.path.join(inpath, gbw_file)):
                    red_exit(
                        "gbw file specified in {} cannot be found".format(
                            input_file
                        )
                    )

            # Per core memory
            if '%maxcore' in line.lower():
                mem_found = True

                if len(line.split()) != 2:
                    red_exit(
                        "Incorrect %maxcore definition in {}".format(
                            input_file
                        )
                    )

                try:
                    n_mb = int(line.split()[-1])
                except ValueError:
                    red_exit(
                        "Cannot parse per core memory in {}".format(input_file)
                    )

            # Number of cores
            if 'pal nprocs' in line.lower():
                n_cores = int(line.split()[2])
                core_found = True

                if n_cores > max_cores:

                    string = "Warning: Specified number of cores"
                    string += " {:d} in {} exceeds".format(
                        n_cores, input_file
                    )
                    string += " node limit of {:d}".format(max_cores)

                    red_exit(string)

    if not mem_found:
        red_exit("Cannot locate %maxcore definition in {}".format(input_file))

    if not core_found:
        red_exit("Cannot locate %maxcore definition in {}".format(input_file))

    # Check memory doesnt exceed per-core limit
    if n_mb > max_mem / n_cores:

        string = "Warning: Specified per core memory of"
        string += " {:d} MB in {} exceeds".format(
            n_mb, input_file
        )
        string += " node limit of {:.2f} MB".format(max_mem / n_cores)

        red_exit(string)

    return full_path_deps, rel_path_deps


def parse_results_contents(input_file):
    """
    Checks results directory (if it exists) for gbw file

    Parameters
    ----------
    input_file : str
        Name of orca input file

    Returns
    -------
    dict[str: str]
        Names of dependencies (files) which this calculation could use
        key is identifier (gbw), value is file name
        Empty if no results directory found
    """

    dependencies = dict()

    job_name = os.path.splitext(input_file)[0]

    if os.path.isdir(job_name):
        if os.path.isfile("{}/{}.gbw".format(job_name, job_name)):
            dependencies["gbw"] = "{}/{}.gbw".format(job_name, job_name)

    return dependencies


def resolve_deps(deps1: dict, deps2: dict):
    """
    Creates a single list of dependencies from two (overlapping) dicts
    """

    deps = dict()

    if len(deps2):
        for key, val in deps2.items():
            deps[key] = val

    if len(deps1):
        for key, val in deps1.items():
            deps[key] = val

    return deps.values()


def add_core_to_input(input_file: str, n_cores: int) -> None:
    """
    Adds number of cores (NPROCS) definition to specified input file

    Parameters
    ----------
    input_file : str
        Name of orca input file
    n_cores : int
        Number of cores to specify

    Returns
    -------
    None
    """

    found = False

    new_file = "{}_tmp".format(input_file)

    with open(input_file, 'r') as fold:
        with open(new_file, 'w') as fnew:

            # Find line if already exists
            for oline in fold:
                # Number of cores
                if 'pal nprocs' in oline.lower():
                    fnew.write("%PAL NPROCS {:d} END\n".format(n_cores))
                    found = True
                else:
                    fnew.write("{}".format(oline))

            # Add if missing
            if not found:
                fnew.write("\n%PAL NPROCS {:d} END\n".format(n_cores))

    subprocess.call("mv {} {}".format(new_file, input_file), shell=True)

    return



import sys


def red_exit(string):
    return sys.exit('\u001b[31m Error: {} \033[0m'.format(string))


def blue_print(string):
    return print('\u001b[34m{} \033[0m'.format(string))


def yellow_print(string):
    return print('\u001b[43m{} \033[0m'.format(string))


def get_opt_coords(file_name: str) -> tuple[list[str], list[float]]:
    '''
    Extracts coordinates from orca optimisation cycle

    Parameters
    ----------
    file_name: str
        Name of file to check

    Returns
    -------
    list[str]
        Labels
    list[float]
        Coordinates (3,n_atoms)
    bool
        True if stationary point found
    '''

    opt_yn = False

    n_cycles = 0

    with open(file_name, 'r') as f:
        for line in f:
            # Optimisation not finished
            if 'GEOMETRY OPTIMIZATION CYCLE' in line:
                n_cycles += 1
                labels = []
                coords = []
                for _ in range(5):
                    line = next(f)
                while len(line.split()) == 4:
                    labels.append(line.split()[0])
                    coords.append([float(val) for val in line.split()[1:]])
                    line = next(f)
            # Optimisation finished, read again
            if '*** FINAL ENERGY EVALUATION AT THE STATIONARY POINT ***' in line: # noqa
                labels = []
                coords = []
                opt_yn = True
                for _ in range(6):
                    line = next(f)
                while len(line.split()) == 4:
                    labels.append(line.split()[0])
                    coords.append([float(val) for val in line.split()[1:]])
                    line = next(f)

    if n_cycles == 0:
        red_exit(
            'Cannot find optimisation cycle coordinates in {}'.format(
                file_name
            )
        )

    return labels, coords, opt_yn


def get_input_section(file_name: str) -> str:
    '''
    Extracts Input section from orca output file
    '''

    input_str = ''

    with open(file_name, 'r') as f:
        for line in f:
            if 'INPUT FILE' in line:
                for _ in range(3):
                    line = next(f)
                while '****END OF INPUT****' not in line:
                    input_str += '{}'.format(line[line.index('> ') + 2:])
                    line = next(f)

    if not len(input_str):
        red_exit(
            'Cannot find input section in {}'.format(
                file_name
            )
        )

    return input_str
